from container_app_conf import ConfigBase
from container_app_conf.entry.file import FileConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from container_app_conf.source.env_source import EnvSource
from container_app_conf.source.toml_source import TomlSource
from container_app_conf.source.yaml_source import YamlSource

NODE_ROOT = "n26"

MFA_TYPE_APP = "app"
MFA_TYPE_SMS = "sms"


class Config(ConfigBase):

    def __new__(cls, *args, **kwargs):
        if "data_sources" not in kwargs.keys():
            yaml_source = YamlSource("n26")
            toml_source = TomlSource("n26")
            data_sources = [
                EnvSource(),
                yaml_source,
                toml_source
            ]
            kwargs["data_sources"] = data_sources

        return super(Config, cls).__new__(cls, *args, **kwargs)

    USERNAME = StringConfigEntry(
        description="N26 account username",
        example="john.doe@example.com",
        key_path=[
            NODE_ROOT,
            "username"
        ],
        required=True
    )

    PASSWORD = StringConfigEntry(
        description="N26 account password",
        example="$upersecret",
        key_path=[
            NODE_ROOT,
            "password"
        ],
        required=True,
        secret=True
    )

    LOGIN_DATA_STORE_PATH = FileConfigEntry(
        description="File path to store login data",
        example="~/.config/n26/token_data",
        key_path=[
            NODE_ROOT,
            "login_data_store_path"
        ],
        required=False,
        default=None
    )

    MFA_TYPE = StringConfigEntry(
        description="Multi-Factor-Authentication type to use",
        example=MFA_TYPE_APP,
        key_path=[
            NODE_ROOT,
            "mfa_type"
        ],
        regex="^({})$".format("|".join([MFA_TYPE_APP, MFA_TYPE_SMS])),
        default=MFA_TYPE_APP
    )



import functools
import logging
import webbrowser
from datetime import datetime, timezone
from typing import Tuple

import click
from requests import HTTPError
from tabulate import tabulate

import n26.api as api
from n26.const import AMOUNT, CURRENCY, REFERENCE_TEXT, ATM_WITHDRAW, CARD_STATUS_ACTIVE, DATETIME_FORMATS

LOGGER = logging.getLogger(__name__)

API_CLIENT = api.Api()

JSON_OUTPUT = False
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def auth_decorator(func: callable):
    """
    Decorator ensuring authentication before making api requests
    :param func: function to patch
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        new_auth = False
        try:
            API_CLIENT.refresh_authentication()
        except HTTPError as http_error:
            if http_error.response.status_code != 401:
                raise http_error
            new_auth = True
        except AssertionError:
            new_auth = True

        if new_auth:
            hint = click.style("Initiating authentication flow, please check your phone to approve login.", fg="yellow")
            click.echo(hint)

            API_CLIENT.authenticate()

            success = click.style("Authentication successful :)", fg="green")
            click.echo(success)

        return func(*args, **kwargs)

    return wrapper


# Cli returns command line requests
@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("-json", default=False, type=bool, is_flag=True)
@click.version_option()
def cli(json: bool):
    """Interact with the https://n26.com API via the command line."""
    global JSON_OUTPUT
    JSON_OUTPUT = json


@cli.command()
@auth_decorator
def addresses():
    """ Show account addresses """
    addresses_data = API_CLIENT.get_addresses().get('data')
    if JSON_OUTPUT:
        _print_json(addresses_data)
        return

    headers = ['Type', 'Country', 'City', 'Zip code', 'Street', 'Number',
               'Address line 1', 'Address line 2',
               'Created', 'Updated']
    keys = ['type', 'countryName', 'cityName', 'zipCode', 'streetName', 'houseNumberBlock',
            'addressLine1', 'addressLine2',
            _datetime_extractor('created'), _datetime_extractor('updated')]
    table = _create_table_from_dict(headers, keys, addresses_data, numalign='right')
    click.echo(table)


@cli.command()
@auth_decorator
def info():
    """ Get account information """
    account_info = API_CLIENT.get_account_info()
    if JSON_OUTPUT:
        _print_json(account_info)
        return

    lines = [
        ["Name:", "{} {}".format(account_info.get('firstName'), account_info.get('lastName'))],
        ["Email:", account_info.get('email')],
        ["Gender:", account_info.get('gender')],
        ["Nationality:", account_info.get('nationality')],
        ["Phone:", account_info.get('mobilePhoneNumber')]
    ]

    text = tabulate(lines, [], tablefmt="plain", colalign=["right", "left"])

    click.echo(text)


@cli.command()
@auth_decorator
def status():
    """ Get account statuses """
    account_statuses = API_CLIENT.get_account_statuses()
    if JSON_OUTPUT:
        _print_json(account_statuses)
        return

    lines = [
        ["Account created:", _timestamp_ms_to_date(account_statuses.get('created'))],
        ["Account updated:", _timestamp_ms_to_date(account_statuses.get('updated'))],
        ["Account closed:", account_statuses.get('accountClosed')],
        ["Card activation completed:", _timestamp_ms_to_date(account_statuses.get('cardActivationCompleted'))],
        ["Card issued:", _timestamp_ms_to_date(account_statuses.get('cardIssued'))],
        ["Core data updated:", _timestamp_ms_to_date(account_statuses.get('coreDataUpdated'))],
        ["Email validation initiated:", _timestamp_ms_to_date(account_statuses.get('emailValidationInitiated'))],
        ["Email validation completed:", _timestamp_ms_to_date(account_statuses.get('emailValidationCompleted'))],
        ["First incoming transaction:", _timestamp_ms_to_date(account_statuses.get('firstIncomingTransaction'))],
        ["Flex account:", account_statuses.get('flexAccount')],
        ["Flex account confirmed:", _timestamp_ms_to_date(account_statuses.get('flexAccountConfirmed'))],
        ["Is deceased:", account_statuses.get('isDeceased')],
        ["Pairing State:", account_statuses.get('pairingState')],
        ["Phone pairing initiated:", _timestamp_ms_to_date(account_statuses.get('phonePairingInitiated'))],
        ["Phone pairing completed:", _timestamp_ms_to_date(account_statuses.get('phonePairingCompleted'))],
        ["Pin definition completed:", _timestamp_ms_to_date(account_statuses.get('pinDefinitionCompleted'))],
        ["Product selection completed:", _timestamp_ms_to_date(account_statuses.get('productSelectionCompleted'))],
        ["Signup step:", account_statuses.get('signupStep')],
        ["Single step signup:", _timestamp_ms_to_date(account_statuses.get('singleStepSignup'))],
        ["Unpairing process status:", account_statuses.get('unpairingProcessStatus')],
    ]

    text = tabulate(lines, [], tablefmt="plain", colalign=["right", "left"])

    click.echo(text)


@cli.command()
@auth_decorator
def balance():
    """ Show account balance """
    balance_data = API_CLIENT.get_balance()
    if JSON_OUTPUT:
        _print_json(balance_data)
        return

    amount = balance_data.get('availableBalance')
    currency = balance_data.get('currency')
    click.echo("{} {}".format(amount, currency))


@cli.command()
def browse():
    """ Browse on the web https://app.n26.com/ """
    webbrowser.open('https://app.n26.com/')


@cli.command()
@auth_decorator
def spaces():
    """ Show spaces """
    spaces_data = API_CLIENT.get_spaces()["spaces"]
    if JSON_OUTPUT:
        _print_json(spaces_data)
        return

    lines = []
    for i, space in enumerate(spaces_data):
        line = []
        available_balance = space['balance']['availableBalance']
        currency = space['balance']['currency']
        name = space['name']

        line.append(name)
        line.append("{} {}".format(available_balance, currency))

        if 'goal' in space:
            goal = space['goal']['amount']
            percentage = available_balance / goal

            line.append("{} {}".format(goal, currency))
            line.append('{:.2%}'.format(percentage))
        else:
            line.append("-")
            line.append("-")

        lines.append(line)

    headers = ['Name', 'Balance', 'Goal', 'Progress']
    text = tabulate(lines, headers, colalign=['left', 'right', 'right', 'right'], numalign='right')

    click.echo(text)


@cli.command()
@auth_decorator
def cards():
    """ Shows a list of cards """
    cards_data = API_CLIENT.get_cards()
    if JSON_OUTPUT:
        _print_json(cards_data)
        return

    headers = ['Id', 'Masked Pan', 'Type', 'Design', 'Status', 'Activated', 'Pin defined', 'Expires']
    keys = [
        'id',
        'maskedPan',
        'cardType',
        'design',
        lambda x: "active" if (x.get('status') == CARD_STATUS_ACTIVE) else x.get('status'),
        _datetime_extractor('cardActivated'),
        _datetime_extractor('pinDefined'),
        _datetime_extractor('expirationDate', date_only=True),
    ]
    text = _create_table_from_dict(headers=headers, value_functions=keys, data=cards_data, numalign='right')

    click.echo(text.strip())


@cli.command()
@click.option('--card', default=None, type=str, help='ID of the card to block. Omitting this will block all cards.')
@auth_decorator
def card_block(card: str):
    """ Blocks the card/s """
    if card:
        card_ids = [card]
    else:
        card_ids = [card['id'] for card in API_CLIENT.get_cards()]

    for card_id in card_ids:
        API_CLIENT.block_card(card_id)
        click.echo('Blocked card: ' + card_id)


@cli.command()
@click.option('--card', default=None, type=str, help='ID of the card to unblock. Omitting this will unblock all cards.')
@auth_decorator
def card_unblock(card: str):
    """ Unblocks the card/s """
    if card:
        card_ids = [card]
    else:
        card_ids = [card['id'] for card in API_CLIENT.get_cards()]

    for card_id in card_ids:
        API_CLIENT.unblock_card(card_id)
        click.echo('Unblocked card: ' + card_id)


@cli.command()
@auth_decorator
def limits():
    """ Show n26 account limits """
    _limits()


@cli.command()
@click.option('--withdrawal', default=None, type=int, help='Daily withdrawal limit.')
@click.option('--payment', default=None, type=int, help='Daily payment limit.')
@auth_decorator
def set_limits(withdrawal: int, payment: int):
    """ Set n26 account limits """
    API_CLIENT.set_account_limits(withdrawal, payment)
    _limits()


def _limits():
    limits_data = API_CLIENT.get_account_limits()

    if JSON_OUTPUT:
        _print_json(limits_data)
        return

    headers = ['Name', 'Amount', 'Country List']
    keys = ['limit', 'amount', 'countryList']
    text = _create_table_from_dict(headers=headers, value_functions=keys, data=limits_data, numalign='right')

    click.echo(text)


@cli.command()
@auth_decorator
def contacts():
    """ Show your n26 contacts """
    contacts_data = API_CLIENT.get_contacts()

    if JSON_OUTPUT:
        _print_json(contacts_data)
        return

    headers = ['Id', 'Name', 'IBAN']
    keys = ['id', 'name', 'subtitle']
    text = _create_table_from_dict(headers=headers, value_functions=keys, data=contacts_data, numalign='right')

    click.echo(text.strip())


@cli.command()
@auth_decorator
def statements():
    """ Show your n26 statements  """
    statements_data = API_CLIENT.get_statements()

    if JSON_OUTPUT:
        _print_json(statements_data)
        return

    headers = ['Id', 'Url', 'Visible TS', 'Month', 'Year']
    keys = ['id', 'url', 'visibleTS', 'month', 'year']
    text = _create_table_from_dict(headers=headers, value_functions=keys, data=statements_data, numalign='right')

    click.echo(text.strip())


@cli.command()
@click.option('--categories', default=None, type=str,
              help='Comma separated list of category IDs.')
@click.option('--pending', default=None, type=bool,
              help='Whether to show only pending transactions.')
@click.option('--from', 'param_from', default=None, type=click.DateTime(DATETIME_FORMATS),
              help='Start time limit for statistics.')
@click.option('--to', 'param_to', default=None, type=click.DateTime(DATETIME_FORMATS),
              help='End time limit for statistics.')
@click.option('--text-filter', default=None, type=str, help='Text filter.')
@click.option('--limit', default=None, type=click.IntRange(1, 10000), help='Limit transaction output.')
@auth_decorator
def transactions(categories: str, pending: bool, param_from: datetime or None, param_to: datetime or None,
                 text_filter: str, limit: int):
    """ Show transactions (default: 5) """
    if not JSON_OUTPUT and not pending and not param_from and not limit:
        limit = 5
        click.echo(click.style("Output is limited to {} entries.".format(limit), fg="yellow"))

    from_timestamp, to_timestamp = _parse_from_to_timestamps(param_from, param_to)
    transactions_data = API_CLIENT.get_transactions(from_time=from_timestamp, to_time=to_timestamp,
                                                    limit=limit, pending=pending, text_filter=text_filter,
                                                    categories=categories)

    if JSON_OUTPUT:
        _print_json(transactions_data)
        return

    lines = []
    for i, transaction in enumerate(transactions_data):
        amount = transaction.get(AMOUNT, 0)
        currency = transaction.get(CURRENCY, None)

        if amount < 0:
            sender_name = "You"
            sender_iban = ""
            recipient_name = transaction.get('merchantName', transaction.get('partnerName', ''))
            recipient_iban = transaction.get('partnerIban', '')
        else:
            sender_name = transaction.get('partnerName', '')
            sender_iban = transaction.get('partnerIban', '')
            recipient_name = "You"
            recipient_iban = ""

        recurring = transaction.get('recurring', '')

        if transaction['type'] == ATM_WITHDRAW:
            message = "ATM Withdrawal"
        else:
            message = transaction.get(REFERENCE_TEXT)

        lines.append([
            _datetime_extractor('visibleTS')(transaction),
            "{} {}".format(amount, currency),
            "{}\n{}".format(sender_name, sender_iban),
            "{}\n{}".format(recipient_name, recipient_iban),
            _insert_newlines(message),
            recurring
        ])

    headers = ['Date', 'Amount', 'From', 'To', 'Message', 'Recurring']
    text = tabulate(lines, headers, numalign='right')

    click.echo(text.strip())


@cli.command("standing-orders")
@auth_decorator
def standing_orders():
    """Show your standing orders"""
    standing_orders_data = API_CLIENT.get_standing_orders()

    if JSON_OUTPUT:
        _print_json(standing_orders_data)
        return

    headers = ['To',
               'Amount',
               'Frequency',
               'Until',
               'Every',
               'First execution', 'Next execution',
               'Executions',
               'Created', 'Updated']
    values = ['partnerName',
              lambda x: "{} {}".format(x.get('amount'), x.get('currencyCode').get('currencyCode')),
              'executionFrequency',
              _datetime_extractor('stopTS', date_only=True),
              _day_of_month_extractor('initialDayOfMonth'),
              _datetime_extractor('firstExecutingTS', date_only=True),
              _datetime_extractor('nextExecutingTS', date_only=True),
              'executionCounter',
              _datetime_extractor('created'),
              _datetime_extractor('updated')]
    text = _create_table_from_dict(headers, value_functions=values, data=standing_orders_data['data'])

    click.echo(text.strip())


@cli.command()
@click.option('--from', 'param_from', default=None, type=click.DateTime(DATETIME_FORMATS),
              help='Start time limit for statistics.')
@click.option('--to', 'param_to', default=None, type=click.DateTime(DATETIME_FORMATS),
              help='End time limit for statistics.')
@auth_decorator
def statistics(param_from: datetime or None, param_to: datetime or None):
    """Show your n26 statistics"""

    from_timestamp, to_timestamp = _parse_from_to_timestamps(param_from, param_to)
    statements_data = API_CLIENT.get_statistics(from_time=from_timestamp, to_time=to_timestamp)

    if JSON_OUTPUT:
        _print_json(statements_data)
        return

    text = "From: %s\n" % (_timestamp_ms_to_date(statements_data["from"]))
    text += "To:   %s\n\n" % (_timestamp_ms_to_date(statements_data["to"]))

    headers = ['Total', 'Income', 'Expense', '#IncomeCategories', '#ExpenseCategories']
    values = ['total', 'totalIncome', 'totalExpense',
              lambda x: len(x.get('incomeItems')),
              lambda x: len(x.get('expenseItems'))]

    text += _create_table_from_dict(headers, value_functions=values, data=[statements_data])

    text += "\n\n"

    headers = ['Category', 'Income', 'Expense', 'Total']
    keys = ['id', 'income', 'expense', 'total']
    text += _create_table_from_dict(headers, keys, statements_data["items"], numalign='right')

    click.echo(text.strip())


def _print_json(data: dict or list):
    """
    Pretty-Prints the given object to the  console
    :param data: data to print
    """
    import json
    json_data = json.dumps(data, indent=2)
    click.echo(json_data)


def _parse_from_to_timestamps(param_from: datetime or None, param_to: datetime or None) -> Tuple[int, int]:
    """
    Parses cli datetime inputs for "from" and "to" parameters
    :param param_from: "from" input
    :param param_to: "to" input
    :return: timestamps ready to be used by the api
    """
    from_timestamp = None
    to_timestamp = None
    if param_from is not None:
        from_timestamp = int(param_from.timestamp() * 1000)
        if param_to is None:
            # if --from is set, --to must also be set
            param_to = datetime.utcnow()
    if param_to is not None:
        if param_from is None:
            # if --to is set, --from must also be set
            from_timestamp = 1
        to_timestamp = int(param_to.timestamp() * 1000)

    return from_timestamp, to_timestamp


def _timestamp_ms_to_date(epoch_ms: int) -> datetime or None:
    """
    Convert millisecond timestamp to UTC datetime.

    :param epoch_ms: milliseconds since 1970 in CET
    :return: a UTC datetime object
    """
    if epoch_ms:
        return datetime.fromtimestamp(epoch_ms / 1000, timezone.utc)


def _create_table_from_dict(headers: list, value_functions: list, data: list, **tabulate_args) -> str:
    """
    Helper function to turn a list of dictionaries into a table.

    Note: This method does NOT work with nested dictionaries and will only inspect top-level keys

    :param headers: the headers to use for the columns
    :param value_functions: function that extracts the value for a given key. Can also be a simple string that
                            will be used as dictionary key.
    :param data: a list of dictionaries containing the data
    :return: a table
    """

    if len(headers) != len(value_functions):
        raise AttributeError("Number of headers does not match number of keys!")

    lines = []
    if isinstance(data, list):
        for dictionary in data:
            line = []
            for value_function in value_functions:
                if callable(value_function):
                    value = value_function(dictionary)
                    line.append(value)
                else:
                    # try to use is as a dict key
                    line.append(dictionary.get(str(value_function)))

            lines.append(line)

    return tabulate(tabular_data=lines, headers=headers, **tabulate_args)


def _datetime_extractor(key: str, date_only: bool = False):
    """
    Helper function to extract a datetime value from a dict
    :param key: the dictionary key used to access the value
    :param date_only: removes the time from the output
    :return: an extractor function
    """

    if date_only:
        fmt = "%x"
    else:
        fmt = "%x %X"

    def extractor(dictionary: dict):
        value = dictionary.get(key)
        time = _timestamp_ms_to_date(value)
        if time is None:
            return None
        else:
            time = time.astimezone()
            return time.strftime(fmt)

    return extractor


def _day_of_month_extractor(key: str):
    def extractor(dictionary: dict):
        value = dictionary.get(key)
        if value is None:
            return None
        else:
            import inflect
            engine = inflect.engine()
            return engine.ordinal(value)

    return extractor


def _insert_newlines(text: str, n=40):
    """
    Inserts a newline into the given text every n characters.
    :param text: the text to break
    :param n:
    :return:
    """
    if not text:
        return ""

    lines = []
    for i in range(0, len(text), n):
        lines.append(text[i:i + n])
    return '\n'.join(lines)


if __name__ == '__main__':
    cli()



def create_request_url(url: str, params: dict = None):
    """
    Adds query params to the given url

    :param url: the url to extend
    :param params: query params as a keyed dictionary
    :return: the url including the given query params
    """

    if params:
        first_param = True
        for k, v in sorted(params.items(), key=lambda entry: entry[0]):
            if not v:
                # skip None values
                continue

            if first_param:
                url += '?'
                first_param = False
            else:
                url += '&'

            url += "%s=%s" % (k, v)

    return url



import json
import logging
import os
import time
from pathlib import Path

import click
import requests
from requests import HTTPError
from tenacity import retry, stop_after_delay, wait_fixed

from n26.config import Config, MFA_TYPE_SMS
from n26.const import DAILY_WITHDRAWAL_LIMIT, DAILY_PAYMENT_LIMIT
from n26.util import create_request_url

LOGGER = logging.getLogger(__name__)

BASE_URL_DE = 'https://api.tech26.de'
BASE_URL_GLOBAL = 'https://api.tech26.global'
BASIC_AUTH_HEADERS = {"Authorization": "Basic bXktdHJ1c3RlZC13ZHBDbGllbnQ6c2VjcmV0"}
USER_AGENT = ("Mozilla/5.0 (X11; Linux x86_64) "
              "AppleWebKit/537.36 (KHTML, like Gecko) "
              "Chrome/59.0.3071.86 Safari/537.36")

GET = "get"
POST = "post"

EXPIRATION_TIME_KEY = "expiration_time"
ACCESS_TOKEN_KEY = "access_token"
REFRESH_TOKEN_KEY = "refresh_token"

GRANT_TYPE_PASSWORD = "password"
GRANT_TYPE_REFRESH_TOKEN = "refresh_token"


class Api(object):
    """
    Api class can be imported as a library in order to use it within applications
    """

    def __init__(self, cfg: Config = None):
        """
        Constructor accepting None to maintain backward compatibility

        :param cfg: configuration object
        """
        if not cfg:
            cfg = Config()
        self.config = cfg
        self._token_data = {}

    @property
    def token_data(self) -> dict:
        if self.config.LOGIN_DATA_STORE_PATH.value is None:
            return self._token_data
        else:
            return self._read_token_file(self.config.LOGIN_DATA_STORE_PATH.value)

    @token_data.setter
    def token_data(self, data: dict):
        if self.config.LOGIN_DATA_STORE_PATH.value is None:
            self._token_data = data
        else:
            self._write_token_file(data, self.config.LOGIN_DATA_STORE_PATH.value)

    @staticmethod
    def _read_token_file(path: str) -> dict:
        """
        :return: the stored token data or an empty dict
        """
        LOGGER.debug("Reading token data from {}".format(path))
        path = Path(path).expanduser().resolve()
        if not path.exists():
            return {}

        if not path.is_file():
            raise IsADirectoryError("File path exists and is not a file: {}".format(path))

        if path.stat().st_size <= 0:
            # file is empty
            return {}

        with open(path, "r") as file:
            return json.loads(file.read())

    @staticmethod
    def _write_token_file(token_data: dict, path: str):
        LOGGER.debug("Writing token data to {}".format(path))
        path = Path(path).expanduser().resolve()

        # delete existing file if permissions don't match or file size is abnormally small
        if path.exists() and (path.stat().st_mode != 0o100600 or path.stat().st_size < 10):
            path.unlink()

        path.parent.mkdir(parents=True, exist_ok=True, mode=0o700)
        with os.fdopen(os.open(path, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as file:
            file.seek(0)
            file.write(json.dumps(token_data, indent=2))
            file.truncate()

    # IDEA: @get_token decorator
    def get_account_info(self) -> dict:
        """
        Retrieves basic account information
        """
        return self._do_request(GET, BASE_URL_DE + '/api/me')

    def get_account_statuses(self) -> dict:
        """
        Retrieves additional account information
        """
        return self._do_request(GET, BASE_URL_DE + '/api/me/statuses')

    def get_addresses(self) -> dict:
        """
        Retrieves a list of addresses of the account owner
        """
        return self._do_request(GET, BASE_URL_DE + '/api/addresses')

    def get_balance(self) -> dict:
        """
        Retrieves the current balance
        """
        return self._do_request(GET, BASE_URL_DE + '/api/accounts')

    def get_spaces(self) -> dict:
        """
        Retrieves a list of all spaces
        """
        return self._do_request(GET, BASE_URL_DE + '/api/spaces')

    def barzahlen_check(self) -> dict:
        return self._do_request(GET, BASE_URL_DE + '/api/barzahlen/check')

    def get_cards(self):
        """
        Retrieves a list of all cards
        """
        return self._do_request(GET, BASE_URL_DE + '/api/v2/cards')

    def get_account_limits(self) -> list:
        """
        Retrieves a list of all active account limits
        """
        return self._do_request(GET, BASE_URL_DE + '/api/settings/account/limits')

    def set_account_limits(self, daily_withdrawal_limit: int = None, daily_payment_limit: int = None) -> None:
        """
        Sets account limits

        :param daily_withdrawal_limit: daily withdrawal limit
        :param daily_payment_limit: daily payment limit
        """
        if daily_withdrawal_limit is not None:
            self._do_request(POST, BASE_URL_DE + '/api/settings/account/limits', json={
                "limit": DAILY_WITHDRAWAL_LIMIT,
                "amount": daily_withdrawal_limit
            })

        if daily_payment_limit is not None:
            self._do_request(POST, BASE_URL_DE + '/api/settings/account/limits', json={
                "limit": DAILY_PAYMENT_LIMIT,
                "amount": daily_payment_limit
            })

    def get_contacts(self):
        """
        Retrieves a list of all contacts
        """
        return self._do_request(GET, BASE_URL_DE + '/api/smrt/contacts')

    def get_standing_orders(self) -> dict:
        """
        Get a list of standing orders
        """
        return self._do_request(GET, BASE_URL_DE + '/api/transactions/so')

    def get_transactions(self, from_time: int = None, to_time: int = None, limit: int = 20, pending: bool = None,
                         categories: str = None, text_filter: str = None, last_id: str = None) -> dict:
        """
        Get a list of transactions.

        Note that some parameters can not be combined in a single request (like text_filter and pending) and
        will result in a bad request (400) error.

        :param from_time: earliest transaction time as a Timestamp > 0 - milliseconds since 1970 in CET
        :param to_time: latest transaction time as a Timestamp > 0 - milliseconds since 1970 in CET
        :param limit: Limit the number of transactions to return to the given amount - default 20 as the n26 API returns
        only the last 20 transactions by default
        :param pending: show only pending transactions
        :param categories: Comma separated list of category IDs
        :param text_filter: Query string to search for
        :param last_id: ??
        :return: list of transactions
        """
        if pending and limit:
            # pending does not support limit
            limit = None

        return self._do_request(GET, BASE_URL_DE + '/api/smrt/transactions', {
            'from': from_time,
            'to': to_time,
            'limit': limit,
            'pending': pending,
            'categories': categories,
            'textFilter': text_filter,
            'lastId': last_id
        })

    def get_transactions_limited(self, limit: int = 5) -> dict:
        import warnings
        warnings.warn(
            "get_transactions_limited is deprecated, use get_transactions(limit=5) instead",
            DeprecationWarning
        )
        return self.get_transactions(limit=limit)

    def get_statements(self) -> list:
        """
        Retrieves a list of all statements
        """
        return self._do_request(GET, BASE_URL_DE + '/api/statements')

    def block_card(self, card_id: str) -> dict:
        """
        Blocks a card.
        If the card is already blocked this will have no effect.

        :param card_id: the id of the card to block
        :return: some info about the card (not including it's blocked state... thanks n26!)
        """
        return self._do_request(POST, BASE_URL_DE + '/api/cards/%s/block' % card_id)

    def unblock_card(self, card_id: str) -> dict:
        """
        Unblocks a card.
        If the card is already unblocked this will have no effect.

        :param card_id: the id of the card to block
        :return: some info about the card (not including it's unblocked state... thanks n26!)
        """
        return self._do_request(POST, BASE_URL_DE + '/api/cards/%s/unblock' % card_id)

    def get_savings(self) -> dict:
        return self._do_request(GET, BASE_URL_DE + '/api/hub/savings/accounts')

    def get_statistics(self, from_time: int = 0, to_time: int = int(time.time()) * 1000) -> dict:
        """
        Get statistics in a given time frame

        :param from_time: Timestamp - milliseconds since 1970 in CET
        :param to_time: Timestamp - milliseconds since 1970 in CET
        """

        if not from_time:
            from_time = 0

        if not to_time:
            to_time = int(time.time()) * 1000

        return self._do_request(GET, BASE_URL_DE + '/api/smrt/statistics/categories/%s/%s' % (from_time, to_time))

    def get_available_categories(self) -> list:
        return self._do_request(GET, BASE_URL_DE + '/api/smrt/categories')

    def get_invitations(self) -> list:
        return self._do_request(GET, BASE_URL_DE + '/api/aff/invitations')

    def _do_request(self, method: str = GET, url: str = "/", params: dict = None,
                    json: dict = None) -> list or dict or None:
        """
        Executes a http request based on the given parameters

        :param method: the method to use (GET, POST)
        :param url: the url to use
        :param params: query parameters that will be appended to the url
        :param json: request body
        :return: the response parsed as a json
        """
        access_token = self.get_token()
        headers = {'Authorization': 'Bearer {}'.format(access_token)}

        url = create_request_url(url, params)

        if method is GET:
            response = requests.get(url, headers=headers, json=json)
        elif method is POST:
            response = requests.post(url, headers=headers, json=json)
        else:
            raise ValueError("Unsupported method: {}".format(method))

        response.raise_for_status()
        # some responses do not return data so we just ignore the body in that case
        if len(response.content) > 0:
            return response.json()

    def is_authenticated(self) -> bool:
        """
        :return: whether valid token data exists
        """
        return self._validate_token(self.token_data)

    def authenticate(self):
        """
        Starts a new authentication flow with the N26 servers.

        This method requires user interaction to approve a 2FA request.
        Therefore you should make sure if you can bypass this
        by refreshing or reusing an existing token by calling is_authenticated()
        and refresh_authentication() respectively.

        :raises PermissionError: if the token is invalid even after the refresh
        """
        LOGGER.debug("Requesting token for username: {}".format(self.config.USERNAME.value))
        token_data = self._request_token(self.config.USERNAME.value, self.config.PASSWORD.value)

        # add expiration time to expiration in _validate_token()
        token_data[EXPIRATION_TIME_KEY] = time.time() + token_data["expires_in"]

        # if it's still not valid, raise an exception
        if not self._validate_token(token_data):
            raise PermissionError("Unable to request authentication token")

        # save token data
        self.token_data = token_data

    def refresh_authentication(self):
        """
        Refreshes an existing authentication using a (possibly expired) token.
        :raises AssertionError: if no existing token data was found
        :raises PermissionError: if the token is invalid even after the refresh
        """
        token_data = self.token_data
        if REFRESH_TOKEN_KEY in token_data:
            LOGGER.debug("Trying to refresh existing token")
            refresh_token = token_data[REFRESH_TOKEN_KEY]
            token_data = self._refresh_token(refresh_token)
        else:
            raise AssertionError("Cant refresh token since no existing token data was found. "
                                 "Please initiate a new authentication instead.")

        # add expiration time to expiration in _validate_token()
        token_data[EXPIRATION_TIME_KEY] = time.time() + token_data["expires_in"]

        # if it's still not valid, raise an exception
        if not self._validate_token(token_data):
            raise PermissionError("Unable to refresh authentication token")

        # save token data
        self.token_data = token_data

    def get_token(self):
        """
        Returns the access token to use for api authentication.
        If a token has been requested before it will be reused if it is still valid.
        If the previous token has expired it will be refreshed.
        If no token has been requested it will be requested from the server.

        :return: the access token
        """
        new_auth = False
        if not self._validate_token(self.token_data):
            try:
                self.refresh_authentication()
            except HTTPError as http_error:
                if http_error.response.status_code != 401:
                    raise http_error
                new_auth = True
            except AssertionError:
                new_auth = True

        if new_auth:
            self.authenticate()

        return self.token_data[ACCESS_TOKEN_KEY]

    def _request_token(self, username: str, password: str):
        """
        Request an authentication token from the server
        :return: the token or None if the response did not contain a token
        """
        mfa_token = self._initiate_authentication_flow(username, password)
        self._request_mfa_approval(mfa_token)
        return self._complete_authentication_flow(mfa_token)

    @staticmethod
    def _initiate_authentication_flow(username: str, password: str) -> str:
        LOGGER.debug("Requesting authentication flow for user {}".format(username))
        values_token = {
            "grant_type": GRANT_TYPE_PASSWORD,
            "username": username,
            "password": password
        }
        # TODO: Seems like the user-agent is not necessary but might be a good idea anyway
        response = requests.post(BASE_URL_GLOBAL + "/oauth/token", data=values_token, headers=BASIC_AUTH_HEADERS)
        if response.status_code != 403:
            raise ValueError("Unexpected response for initial auth request: {}".format(response.text))

        response_data = response.json()
        if response_data.get("error", "") == "mfa_required":
            return response_data["mfaToken"]
        else:
            raise ValueError("Unexpected response data")

    @staticmethod
    def _refresh_token(refresh_token: str):
        """
        Refreshes an authentication token
        :param refresh_token: the refresh token issued by the server when requesting a token
        :return: the refreshed token data
        """
        LOGGER.debug("Requesting token refresh using refresh_token {}".format(refresh_token))
        values_token = {
            'grant_type': GRANT_TYPE_REFRESH_TOKEN,
            'refresh_token': refresh_token,
        }

        response = requests.post(BASE_URL_GLOBAL + '/oauth/token', data=values_token, headers=BASIC_AUTH_HEADERS)
        response.raise_for_status()
        return response.json()

    def _request_mfa_approval(self, mfa_token: str):
        LOGGER.debug("Requesting MFA approval using mfa_token {}".format(mfa_token))
        mfa_data = {
            "mfaToken": mfa_token
        }

        if self.config.MFA_TYPE.value == MFA_TYPE_SMS:
            mfa_data['challengeType'] = "otp"
        else:
            mfa_data['challengeType'] = "oob"

        response = requests.post(
            BASE_URL_DE + "/api/mfa/challenge",
            json=mfa_data,
            headers={
                **BASIC_AUTH_HEADERS,
                "User-Agent": USER_AGENT,
                "Content-Type": "application/json"
            })
        response.raise_for_status()

    @retry(wait=wait_fixed(5), stop=stop_after_delay(60))
    def _complete_authentication_flow(self, mfa_token: str) -> dict:
        LOGGER.debug("Completing authentication flow for mfa_token {}".format(mfa_token))
        mfa_response_data = {
            "mfaToken": mfa_token
        }

        if self.config.MFA_TYPE.value == MFA_TYPE_SMS:
            mfa_response_data['grant_type'] = "mfa_otp"

            hint = click.style("Enter the 6 digit SMS OTP code", fg="yellow")
            mfa_response_data['otp'] = click.prompt(hint, type=int)
        else:
            mfa_response_data['grant_type'] = "mfa_oob"

        response = requests.post(BASE_URL_DE + "/oauth/token", data=mfa_response_data, headers=BASIC_AUTH_HEADERS)
        response.raise_for_status()
        tokens = response.json()
        return tokens

    @staticmethod
    def _validate_token(token_data: dict):
        """
        Checks if a token is valid
        :param token_data: the token data to check
        :return: true if valid, false otherwise
        """
        if EXPIRATION_TIME_KEY not in token_data:
            # there was a problem adding the expiration_time property
            return False
        elif time.time() >= token_data[EXPIRATION_TIME_KEY]:
            # token has expired
            return False

        return ACCESS_TOKEN_KEY in token_data and token_data[ACCESS_TOKEN_KEY]



import inspect
import os

from setuptools import setup

__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))


def read_version(package):
    with open(os.path.join(package, '__init__.py'), 'r') as fd:
        for line in fd:
            if line.startswith('__version__ = '):
                return line.split()[-1].strip().strip("'")


def get_install_requirements(path):
    content = open(os.path.join(__location__, path)).read()
    return [req for req in content.split('\\n') if req != '']


def read(fname):
    return open(os.path.join(__location__, fname)).read()


VERSION = read_version('n26')

setup(
    description='API and command line tools to interact with the https://n26.com/ API',
    long_description='API and command line tools to interact with the https://n26.com/ API',
    author='Felix Mueller',
    author_email='felix@s4ku.com',
    url='https://github.com/femueller/python-n26',
    download_url='https://github.com/femueller/python-n26/tarball/{version}'.format(version=VERSION),
    version=VERSION,
    install_requires=['requests', 'pyyaml', 'click', 'tabulate', 'inflect', 'tenacity', 'container-app-conf'],
    test_requires=['mock', 'pytest'],
    packages=[
        'n26'
    ],
    scripts=[],
    name='n26',
    entry_points={
        'console_scripts': ['n26 = n26.cli:cli']
    }
)



"""
transaction types and their meaning
"""
DEBIT_PAYMENT = "AA"
SEPA_WITHDRAW = "DD"
INCOMING_TRANSFER = "CT"
ATM_WITHDRAW = "PT"

"""
transaction item keys
"""
CURRENCY = 'currencyCode'
AMOUNT = 'amount'
REFERENCE_TEXT = 'referenceText'

CARD_STATUS_ACTIVE = 'M_ACTIVE'

DAILY_WITHDRAWAL_LIMIT = 'ATM_DAILY_ACCOUNT'
DAILY_PAYMENT_LIMIT = 'POS_DAILY_ACCOUNT'

DATETIME_FORMATS = [
    '%m/%d/%Y %H:%M:%S',
    '%m/%d/%Y',
    '%Y-%m-%d',
    '%Y-%m-%dT%H:%M:%S',
    '%Y-%m-%d %H:%M:%S',
    '%d.%m.%Y',
    '%d.%m.%Y %H:%M:%S'
]
